from typing import Any

from fashionable import UNSET, validate

__all__ = [
    'Arg',
    'ArgError',
    'InvalidArgError',
    'MissingArgError',
]


class Arg:
    __slots__ = ('name', 'type', 'default', 'is_positional', 'is_zipped')

    def __init__(self, name: str, type_: type, default: Any, is_positional: bool, is_zipped: bool):
        self.name = name
        self.type = type_
        self.default = default
        self.is_positional = is_positional
        self.is_zipped = is_zipped

    def validate(self, value: Any) -> Any:
        if value is UNSET:
            if self.default is UNSET:
                raise MissingArgError(self)
            else:
                result = self.default
        else:
            try:
                result = validate(self.type, value)
            except (TypeError, ValueError) as exc:
                raise InvalidArgError(self, value) from exc

        return result


class ArgError(Exception):
    def __init__(self, fmt: str, arg: Arg, value: Any = UNSET):
        super().__init__(fmt.format(arg=arg, value=value))
        self.arg = arg
        self.value = value


class MissingArgError(ArgError):
    def __init__(self, *args, **kwargs):
        super().__init__("missing required argument {arg.name!r} value", *args, **kwargs)


class InvalidArgError(ArgError):
    def __init__(self, *args, **kwargs):
        super().__init__("invalid argument {arg.name!r} value {value!r}", *args, **kwargs)



from typing import Union

from .models import Notification, Request, Response

__all__ = [
    'AnyJsonrpc',
    'Incoming',
    'Outgoing',
]

Incoming = Union[Request, Notification]
Outgoing = Union[Response, Notification]
AnyJsonrpc = Union[Incoming, Outgoing]



from .errors import *
from .jsonrpc import *
from .loggers import *
from .models import *
from .notifier import *
from .types import *

__all__ = [
    *errors.__all__,
    *jsonrpc.__all__,
    *loggers.__all__,
    *models.__all__,
    *notifier.__all__,
    *types.__all__,
]

__version__ = '0.2.2'



from asyncio import iscoroutine
from inspect import getfullargspec
from itertools import chain, zip_longest
from typing import Any, Callable, Dict, Optional, Tuple, Iterable

from fashionable import UNSET

from .arg import Arg, ArgError

__all__ = [
    'ResultError',
    'Route',
]


def _zip_right(*iterables: Iterable, fillvalue: Any = None) -> Iterable:
    return reversed(tuple(zip_longest(*map(reversed, iterables), fillvalue=fillvalue)))


class Route:
    __slots__ = ('func', 'method', 'args', 'result')

    @classmethod
    def from_inspect(cls, func: Callable, method: Optional[str], annotations: Dict[str, type]) -> 'Route':
        if not method:
            method = func.__name__

        spec = getfullargspec(func)
        kwonlydefaults = spec.kwonlydefaults or {}

        args = tuple(
            Arg(name, annotations.get(name, spec.annotations.get(name, Any)), default, is_positional, is_zipped)
            for name, default, is_positional, is_zipped in chain(
                ((n, d, True, False) for n, d in _zip_right(spec.args, spec.defaults or (), fillvalue=UNSET)),
                ((n, d, False, False) for n, d in ((a, kwonlydefaults.get(a, UNSET)) for a in spec.kwonlyargs)),
                ((spec.varargs, UNSET, True, True),) if spec.varargs else (),
                ((spec.varkw, UNSET, False, True),) if spec.varkw else (),
            )
        )

        result_type = annotations.get('result', spec.annotations.get('return'))
        result = Arg('result', result_type, UNSET, False, False) if result_type else None

        return cls(func, method, args, result)

    def __init__(self, func: Callable, method: str, args: Tuple[Arg, ...], result: Optional[Arg]):
        self.func = func
        self.method = method
        self.args = args
        self.result = result

    def _validate(self, params: Any, customs: Dict[type, Any]) -> Tuple[list, dict]:
        list_params = None
        dict_params = None

        if isinstance(params, list):
            list_params = params.copy()
        elif isinstance(params, dict):
            dict_params = params.copy()
        else:
            list_params = [params]

        args = []
        kwargs = {}
        recover_allowed = True

        for arg in self.args:
            if arg.is_zipped:
                if dict_params:
                    for param_name, param_value in dict_params.items():
                        kwargs[param_name] = arg.validate(param_value)
                        recover_allowed = False
                else:
                    for param_value in list_params:
                        args.append(arg.validate(param_value))
                        recover_allowed = False

                continue

            value = customs.get(arg.type, UNSET)

            name = arg.name

            if value is UNSET:
                try:
                    value = arg.validate(
                        dict_params.pop(name, UNSET) if dict_params else list_params.pop(0) if list_params else UNSET
                    )
                except ArgError:
                    if not recover_allowed:
                        raise

                    value = arg.validate(params)

            if arg.is_positional:
                args.append(value)
            else:
                kwargs[name] = value

            recover_allowed = False

        return args, kwargs

    async def call(self, *args, **kwargs) -> Any:
        args, kwargs = self._validate(*args, **kwargs)
        ret = self.func(*args, **kwargs)

        if iscoroutine(ret):
            ret = await ret

        if self.result:
            try:
                ret = self.result.validate(ret)
            except ArgError as err:
                raise ResultError(err.arg, ret) from err

        return ret


class ResultError(ArgError):
    def __init__(self, *args, **kwargs):
        super().__init__("invalid call result value {value!r}", *args, **kwargs)



from asyncio import Future, ensure_future
from typing import Callable, Any

from websockets import WebSocketCommonProtocol as WebSocket

from .models import Notification

__all__ = [
    'Notifier',
]

_Sender = Callable[[WebSocket, Notification], Future]
_Finalizer = Callable[[Future], Any]


class Notifier:
    def __init__(self, ws: WebSocket, sender: _Sender, finalizer: _Finalizer):
        self._ws = ws
        self._sender = sender
        self._finalizer = finalizer
        self._pending = set()

    @property
    def closed(self) -> bool:
        return self._ws.closed

    def _done_callback(self, fut: Future):
        self._finalizer(fut)

        if fut in self._pending:
            self._pending.remove(fut)

    def send(self, notification: Notification) -> Future:
        fut = self._sender(self._ws, notification) if self._ws.open else ensure_future(self._ws.ensure_open())
        fut.add_done_callback(self._done_callback)
        self._pending.add(fut)
        return fut

    def cancel(self):
        for fut in self._pending:
            fut.cancel()



from asyncio import Future, gather, ensure_future, wait, CancelledError, FIRST_COMPLETED
from typing import Optional, Dict, Any

from sanic import Sanic
from sanic.request import Request as SanicRequest
from sanic.response import HTTPResponse
from websockets import WebSocketCommonProtocol as WebSocket

from ._basejsonrpc import BaseJsonrpc
from .._routing import Route
from ..loggers import logger, traffic_logger
from ..models import Notification, Response, Request
from ..notifier import Notifier
from ..types import Outgoing, Incoming

__all__ = [
    'Jsonrpc',
    'SanicJsonrpc',
]


# TODO middleware
class SanicJsonrpc(BaseJsonrpc):
    def _customs(
            self,
            sanic_request: SanicRequest,
            incoming: Incoming,
            ws: Optional[WebSocket] = None,
            notifier: Optional[Notifier] = None
    ) -> Dict[type, Any]:
        return {
            SanicRequest: sanic_request,
            WebSocket: ws,
            Sanic: self.app,
            Request: incoming,
            Notification: incoming,
            Incoming: incoming,
            Notifier: notifier,
        }

    async def _post(self, sanic_request: SanicRequest) -> HTTPResponse:
        incomings = self._parse_messages(sanic_request.body)

        single = not isinstance(incomings, list)

        if single:
            incomings = [incomings]

        responses = []
        futures = []

        for incoming in incomings:
            if isinstance(incoming, Response):
                responses.append(incoming)
                continue

            route = self._route(incoming, is_post=True)

            if not isinstance(route, Route):
                if route:
                    responses.append(route)
                else:
                    logger.info("Unhandled %r", incoming)

                continue

            fut = self._register_call(incoming, route, self._customs(sanic_request, incoming))

            if isinstance(incoming, Request):
                futures.append(fut)

        for response in await gather(*futures):
            responses.append(response)

        body = self._serialize_responses(responses, single)
        content_type = 'application/json' if body else 'text/plain'
        return HTTPResponse(body, 207, content_type=content_type)

    def _ws_outgoing(self, ws: WebSocket, outgoing: Outgoing) -> Future:
        traffic_logger.debug("<-- %r", outgoing)
        return ensure_future(ws.send(self._serialize(dict(outgoing))))

    async def _ws(self, sanic_request: SanicRequest, ws: WebSocket):
        recv = None
        pending = set()
        notifier = Notifier(ws, self._ws_outgoing, self._finalise_future)

        while ws.open:
            if recv not in pending:
                recv = ensure_future(ws.recv())
                pending.add(recv)

            try:
                done, pending = await wait(pending, return_when=FIRST_COMPLETED)
            except CancelledError:
                for fut in pending:
                    self._finalise_future(fut)

                break

            for fut in done:
                result = self._finalise_future(fut)

                if not result:
                    continue

                if isinstance(result, Response):
                    pending.add(self._ws_outgoing(ws, result))
                    continue

                obj = self._parse_json(result)

                if isinstance(obj, Response):
                    pending.add(self._ws_outgoing(ws, obj))
                    continue

                incoming = self._parse_message(obj)

                if isinstance(incoming, Response):
                    pending.add(self._ws_outgoing(ws, incoming))
                    continue

                route = self._route(incoming, is_post=False)

                if not isinstance(route, Route):
                    if route:
                        pending.add(self._ws_outgoing(ws, route))
                    else:
                        logger.info("Unhandled %r", incoming)

                    continue

                fut = self._register_call(incoming, route, self._customs(sanic_request, incoming, ws, notifier))

                if isinstance(incoming, Request):
                    pending.add(fut)

        notifier.cancel()

        for fut in pending:
            fut.cancel()

    def __init__(self, app: Sanic, post_route: Optional[str] = None, ws_route: Optional[str] = None, **kwargs):
        super().__init__(**kwargs)
        self.app = app
        self._processing_task = None
        app.listener('after_server_start')(self._start_processing)
        app.listener('before_server_stop')(self._stop_processing)

        if post_route:
            self.app.add_route(self._post, post_route, methods=frozenset({'POST'}))

        if ws_route:
            self.app.add_websocket_route(self._ws, ws_route)


class Jsonrpc(SanicJsonrpc):
    def __init__(self, *args, **kwargs):
        from warnings import warn
        warn(
            "Class {} has been renamed to {}".format(self.__class__.__name__, self.__class__.__base__.__name__),
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(*args, **kwargs)



from logging import getLogger

__all__ = [
    'access_logger',
    'error_logger',
    'logger',
    'traffic_logger',
]

logger = getLogger(__package__ + '.root')
error_logger = getLogger(__package__ + '.error')
traffic_logger = getLogger(__package__ + '.traffic')
access_logger = getLogger(__package__ + '.access')



from typing import Optional

from fashionable import Attribute, Model, ModelValueError

__all__ = [
    '_Jsonrpc',
]


class _Jsonrpc(Model):
    jsonrpc = Attribute(Optional[str], strict=True, default='2.0')

    def __init__(self, *args, **kwargs):
        if args and args[0] != '2.0':
            args = ('2.0',) + args

        super().__init__(*args, **kwargs)

        if self.jsonrpc != '2.0':
            raise ModelValueError(
                'Invalid %(model)s: invalid attribute jsonrpc: MUST be exactly "2.0"', model=self.__class__.__name__
            )



from .error import *
from .notification import *
from .request import *
from .response import *

__all__ = [
    *error.__all__,
    *notification.__all__,
    *request.__all__,
    *response.__all__,
]



from asyncio import shield, Future, Queue, ensure_future
from time import monotonic
from typing import Any, AnyStr, Callable, Dict, List, Optional, Union, Tuple

from fashionable import ModelError, ModelAttributeError, UNSET
from ujson import dumps, loads

from .._routing import Route, ArgError, ResultError
from ..errors import INTERNAL_ERROR, INVALID_PARAMS, INVALID_REQUEST, METHOD_NOT_FOUND, PARSE_ERROR
from ..loggers import access_logger, error_logger, logger, traffic_logger
from ..models import Error, Notification, Request, Response
from ..types import AnyJsonrpc, Incoming

__all__ = [
    'BaseJsonrpc',
]


class BaseJsonrpc:
    @staticmethod
    def _parse_json(json: AnyStr) -> Union[Dict, List[Dict], Response]:
        try:
            return loads(json)
        except (TypeError, ValueError):
            return Response(error=PARSE_ERROR)

    @staticmethod
    def _parse_message(message: Dict) -> AnyJsonrpc:
        try:
            return Request(**message)
        except (TypeError, ModelError) as err:
            if isinstance(err, ModelAttributeError) and err.kwargs['attr'] == 'id':
                return Notification(**message)

            return Response(error=INVALID_REQUEST)

    def _parse_messages(self, json: AnyStr) -> Union[AnyJsonrpc, List[AnyJsonrpc]]:
        messages = self._parse_json(json)

        if isinstance(messages, Response):
            return messages

        if isinstance(messages, list):
            if not messages:
                return Response(error=INVALID_REQUEST)

            return [self._parse_message(m) for m in messages]

        return self._parse_message(messages)

    def _serialize(self, obj: Any) -> str:
        try:
            return dumps(obj)
        except Exception as err:
            error_logger.error("Failed to serialize object %r: %s", obj, err, exc_info=err)
            return self._serialize(dict(Response(error=INTERNAL_ERROR)))

    def _serialize_responses(self, responses: List[Response], single: bool) -> Optional[str]:
        if not responses:
            return None

        if single:
            return self._serialize(dict(responses[0]))

        return self._serialize([dict(r) for r in responses])

    def _route(self, incoming: Incoming, is_post: bool) -> Optional[Union[Route, Response]]:
        is_request = isinstance(incoming, Request)
        route = self._routes.get((is_post, is_request, incoming.method))

        if route:
            return route

        if is_request:
            return Response(error=METHOD_NOT_FOUND, id=incoming.id)

    def _register_call(self, *args, **kwargs) -> Future:
        fut = shield(self._call(*args, **kwargs))
        self._calls.put_nowait(fut)
        return fut

    async def _call(self, incoming: Incoming, route: Route, customs: Dict[type, Any]) -> Optional[Response]:
        access_log = self.access_log

        if access_log:
            start = monotonic()
        else:
            start = None

        traffic_logger.debug("--> %r", incoming)

        error = UNSET
        result = UNSET

        try:
            ret = await route.call(incoming.params, customs)
        except ResultError as err:
            error_logger.error("%r failed: %s", incoming, err, exc_info=err)
            error = INTERNAL_ERROR
        except ArgError as err:
            logger.debug("Invalid %r: %s", incoming, err)
            error = INVALID_PARAMS
        except Error as err:
            error = err
        except Exception as err:
            error_logger.error("%r failed: %s", incoming, err, exc_info=err)
            error = INTERNAL_ERROR
        else:
            if isinstance(ret, Error):
                error = ret
            else:
                result = ret

        is_request = isinstance(incoming, Request)

        if access_log:
            access_logger.info("", extra={
                'type': incoming.__class__.__name__,
                'method': incoming.method,
                'id': incoming.id if is_request else '',
                'time': '{:.6f}'.format((monotonic() - start) * 1000),
                'error': error.code if error is not UNSET else '',
            })

        if is_request:
            response = Response(result=result, error=error, id=incoming.id)
            traffic_logger.debug("<-- %r", response)
            return response

    @staticmethod
    def _finalise_future(fut: Future) -> Optional[Union[Response, str]]:
        if fut.done():
            err = fut.exception()

            if err:
                error_logger.error("%s", err, exc_info=err)
            else:
                return fut.result()
        else:
            fut.cancel()

    async def _processing(self):
        calls = self._calls

        while True:
            call = await calls.get()
            await call

    async def _start_processing(self, _app, _loop):
        self._calls = Queue()
        self._processing_task = ensure_future(self._processing())

    async def _stop_processing(self, _app, _loop):
        self._processing_task.cancel()

        calls = self._calls

        while not calls.empty():
            await calls.get_nowait()

    def __init__(self, *, access_log: bool = True):
        self.access_log = access_log
        self._routes = {}
        self._calls = None

    def __call__(
            self,
            method_: Optional[str] = None,
            *,
            is_post_: Tuple[bool, ...] = (True, False),
            is_request_: Tuple[bool, ...] = (True, False),
            **annotations: type
    ) -> Callable:
        if isinstance(method_, Callable):
            return self.__call__(is_post_=is_post_, is_request_=is_request_)(method_)

        def deco(func: Callable) -> Callable:
            route = Route.from_inspect(func, method_, annotations)
            self._routes.update({(ip, ir, route.method): route for ip in is_post_ for ir in is_request_})
            return func
        return deco

    def post(self, method_: Optional[str] = None, **annotations: type) -> Callable:
        return self.__call__(method_, is_post_=(True,), **annotations)

    def ws(self, method_: Optional[str] = None, **annotations: type) -> Callable:
        return self.__call__(method_, is_post_=(False,), **annotations)

    def request(self, method_: Optional[str] = None, **annotations: type) -> Callable:
        return self.__call__(method_, is_request_=(True,), **annotations)

    def notification(self, method_: Optional[str] = None, **annotations: type) -> Callable:
        return self.__call__(method_, is_request_=(False,), **annotations)

    def post_request(self, method_: Optional[str] = None, **annotations: type) -> Callable:
        return self.__call__(method_, is_post_=(True,), is_request_=(True,), **annotations)

    def ws_request(self, method_: Optional[str] = None, **annotations: type) -> Callable:
        return self.__call__(method_, is_post_=(False,), is_request_=(True,), **annotations)

    def post_notification(self, method_: Optional[str] = None, **annotations: type) -> Callable:
        return self.__call__(method_, is_post_=(True,), is_request_=(False,), **annotations)

    def ws_notification(self, method_: Optional[str] = None, **annotations: type) -> Callable:
        return self.__call__(method_, is_post_=(False,), is_request_=(False,), **annotations)
